from django.contrib import admin
from kubernetes_manager.models import (
    KubernetesConfigMap,
    KubernetesContainer,
    KubernetesDeployment,
    KubernetesIngress,
    KubernetesJob,
    KubernetesNamespace,
    KubernetesPodTemplate,
    KubernetesService,
    KubernetesVolume,
    KubernetesVolumeMount,
    TargetCluster,
)

models = [
    TargetCluster,
    KubernetesContainer,
    KubernetesPodTemplate,
    KubernetesJob,
    KubernetesDeployment,
    KubernetesService,
    KubernetesIngress,
    KubernetesNamespace,
    KubernetesConfigMap,
    KubernetesVolume,
    KubernetesVolumeMount,
]

for model in models:
    admin.site.register(model)



import os

import factory
import factory.fuzzy
from factory.django import DjangoModelFactory as DMF
from faker import Faker

fake = Faker()

models_path = "kubernetes_manager."

microk8s_config = os.getenv("MICROK8S_CF", None)


class TargetClusterFactory(DMF):
    class Meta:
        model = models_path + "TargetCluster"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-cluster")
    description = fake.sentence()
    api_endpoint = "https://127.0.0.1:16443"
    telemetry_endpoint = "https://127.0.0.1:16443"
    telemetry_source = "p"
    config = microk8s_config


class KubernetesNamespaceFactory(DMF):
    class Meta:
        model = models_path + "KubernetesNamespace"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-ns")
    description = fake.paragraph(nb_sentences=3, variable_nb_sentences=True)
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None
    labels = {"app": fake.word()}
    annotations = None
    api_version = "v1"
    kind = "Namespace"
    exists = False


class KubernetesConfigMapFactory(DMF):
    class Meta:
        model = models_path + "KubernetesConfigMap"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-cm")
    description = fake.sentence()
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None
    kind = "ConfigMap"
    data = {"data": str(factory.fuzzy.FuzzyText(length=12))}
    namespace = factory.SubFactory(KubernetesNamespaceFactory)


class KubernetesVolumeFactory(DMF):
    class Meta:
        model = models_path + "KubernetesVolume"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-vol")
    description = fake.sentence()
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None


class KubernetesVolumeMountFactory(DMF):
    class Meta:
        model = models_path + "KubernetesVolumeMount"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-mount")
    description = fake.sentence()
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None
    mount_path = "/media"
    sub_path = None


class KubernetesContainerFactory(DMF):
    class Meta:
        model = models_path + "KubernetesContainer"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-container")
    description = fake.sentence()
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None
    image_name = factory.fuzzy.FuzzyChoice(["debian", "alpine", "busybox"])
    image_tag = "latest"
    image_pull_policy = "IfNotPresent"
    command = "/bin/sh"
    args = "-c,echo SUCCESS"
    port = factory.fuzzy.FuzzyChoice([80, 8080, 8000])


class KubernetesPodTemplateFactory(DMF):
    class Meta:
        model = models_path + "KubernetesPodTemplate"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-pod")
    description = fake.paragraph(nb_sentences=3, variable_nb_sentences=True)
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None
    labels = {"app": fake.word()}
    annotations = None
    restart_policy = "Always"

    @factory.post_generation
    def containers(self, create, containers, **kwargs):
        if not create:
            # Simple build, do nothing.
            return

        if containers:
            # A list of groups were passed in, use them
            for container in containers:
                self.containers.add(container)

    @factory.post_generation
    def volumes(self, create, volumes, **kwargs):
        if not create:
            # Simple build, do nothing.
            return

        if volumes:
            # A list of groups were passed in, use them
            for volume in volumes:
                self.volumes.add(volume)



class KubernetesDeploymentFactory(DMF):
    class Meta:
        model = models_path + "KubernetesDeployment"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-dep")
    description = fake.paragraph(nb_sentences=3, variable_nb_sentences=True)
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None
    labels = {"app": fake.word()}
    annotations = None
    api_version = "apps/v1"
    kind = "Deployment"
    port = factory.fuzzy.FuzzyChoice([80, 8080, 8000])
    namespace = factory.SubFactory(KubernetesNamespaceFactory)
    kuid = None
    selector = labels
    replicas = 1
    pod_template = factory.SubFactory(KubernetesPodTemplateFactory)


class KubernetesJobFactory(DMF):
    class Meta:
        model = models_path + "KubernetesJob"

    title = factory.fuzzy.FuzzyText(length=8, suffix="-job")
    description = fake.paragraph(nb_sentences=3, variable_nb_sentences=True)
    cluster = factory.SubFactory(TargetClusterFactory)
    config = {"data_is_fake": "true"}
    deployed = None
    removed = None
    labels = {"app": fake.word()}
    annotations = None
    api_version = "apps/v1"
    kind = "Deployment"
    port = factory.fuzzy.FuzzyChoice([80, 8080, 8000])
    namespace = factory.SubFactory(KubernetesNamespaceFactory)
    kuid = None
    pod_template = factory.SubFactory(KubernetesPodTemplateFactory)
    backoff_limit = factory.fuzzy.FuzzyInteger(1, 10)



from django.conf.urls import url
from django.contrib import admin
from django.urls import include, path
from kubernetes_manager import views
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework import permissions, routers

schema_view = get_schema_view(
    openapi.Info(
        title="DKM API",
        default_version="v0.1.0",
        description="Django Kubernetes Manager",
        contact=openapi.Contact(email="bradley@introspectdata.com"),
        license=openapi.License(name="MIT License", url="https://github.com/IntrospectData/Django-Kubernetes-Manager/blob/master/LICENSE"),
    ),
    public=True,
    permission_classes=(permissions.IsAuthenticated,),
)

router = routers.DefaultRouter()
router.register(r"containers", views.KubernetesContainerViewSet)
router.register(r"pods", views.KubernetesPodTemplateViewSet)
router.register(r"deployments", views.KubernetesDeploymentViewSet)
router.register(r"services", views.KubernetesServiceViewSet)
router.register(r"ingresses", views.KubernetesIngressViewSet)
router.register(r"jobs", views.KubernetesJobViewSet)
router.register(r"clusters", views.TargetClusterViewSet)
router.register(r"namespaces", views.KubernetesNamespaceViewSet)
router.register(r"configmaps", views.KubernetesConfigMapViewSet)
router.register(r"volumes", views.KubernetesVolumeViewSet)
router.register(r"mounts", views.KubernetesVolumeMountViewSet)


urlpatterns = [
    path("api/", include(router.urls)),
    path("api/auth/", include("rest_framework.urls", namespace="rest")),
    url(r"^api/swagger(?P<format>\.json|\.yaml)$", schema_view.without_ui(cache_timeout=0), name="schema-json"),
    path("api/swagger/", schema_view.with_ui("swagger", cache_timeout=0), name="schema-swagger-ui"),
    path("api/redoc/", schema_view.with_ui("redoc", cache_timeout=0), name="schema-redoc"),
]



from .model_views import (
    KubernetesConfigMapViewSet,
    KubernetesContainerViewSet,
    KubernetesDeploymentViewSet,
    KubernetesIngressViewSet,
    KubernetesJobViewSet,
    KubernetesNamespaceViewSet,
    KubernetesPodTemplateViewSet,
    KubernetesServiceViewSet,
    KubernetesVolumeMountViewSet,
    KubernetesVolumeViewSet,
    TargetClusterViewSet,
)



import json
from time import sleep

from django.contrib.postgres.fields import JSONField
from django.db import models
from kubernetes_manager.consts import PULL_POLICY, RESTART_POLICY
from kubernetes import client, config

from .base import KubernetesBase, KubernetesMetadataObjBase, KubernetesNetworkingBase
from .mixins import KubernetesTelemetryMixin


class KubernetesNamespace(KubernetesMetadataObjBase):
    """
    KubernetesNamespace
    :type: model
    :description: Holds data related to a Kubernetes namespace.
    :inherits: kubernetes_manager.models.base.KubernetesBase
    :fields: api_version, kind, exists
    """

    api_version = models.CharField(max_length=16, default="v1")
    kind = models.CharField(max_length=16, default="Namespace")
    exists = models.BooleanField(default=False)

    def get_obj(self):
        """
        :description: Generate namespace spec.
        """
        return client.V1Namespace(
            api_version=self.api_version,
            kind=self.kind,
            metadata=client.V1ObjectMeta(name=self.slug, labels=self.labels, annotations=self.annotations),
            spec=client.V1NamespaceSpec(),
        )

    def deploy(self):
        """
        :description: Deploy namespace obj.
        """
        api_instance = self.get_client(API=client.CoreV1Api)
        body = self.get_obj()
        api_response = api_instance.create_namespace(body=body,)
        self.exists = True
        self.save()
        return str(api_response.status)

    def remove(self):
        """
        :description: Delete namespace from cluster.
        """
        api_instance = self.get_client(API=client.CoreV1Api)
        api_response = api_instance.delete_namespace(name=self.slug,)
        self.exists = False
        self.save()
        return str(api_response.status)


class KubernetesVolume(KubernetesBase):
    """
    KubernetesVolume
    :type: model
    :description: Holds data related to a kubernetes volume.
    :inherits: kubernetes_manager.models.base.KubernetesBase
    :fields: *
    """

    def get_obj(self):
        """
        :description: Generate volume spec.
        """
        if self.config.get("configmap"):
            return client.V1Volume(name=self.slug, config_map=client.V1ConfigMapVolumeSource(name=self.config.get("configmap")))
        return client.V1Volume(name=self.slug, empty_dir=client.V1EmptyDirVolumeSource())


class KubernetesVolumeMount(KubernetesBase):
    """
    KubernetesVolumeMount
    :type: model
    :description: Holds data related to a kubernetes volume mount.
    :inherits: kubernetes_manager.models.base.KubernetesBase
    :fields: mount_path, sub_path
    """

    mount_path = models.CharField(max_length=255, default="/media")
    sub_path = models.CharField(max_length=255, default=None, null=True, blank=True)

    def get_obj(self):
        """
        :description: Generate mount spec.
        """
        return client.V1VolumeMount(name=self.slug, mount_path=self.mount_path, sub_path=self.sub_path)


class KubernetesContainer(KubernetesBase):
    """
    KubernetesContainer
    :type: model
    :description: Holds data related to a kubernetes contaienr.
    :inherits: kubernetes_manager.models.base.KubernetesBase
    :fields: image_name, image_tag, image_pull_policy, command, args, port,
        volume_mount
    """

    image_name = models.CharField(max_length=200, db_index=True, help_text="Properly qualified image name.", default="debian")
    image_tag = models.CharField(max_length=100, db_index=True, help_text="Tag name for the image to be used for this job", default="latest")
    image_pull_policy = models.CharField(max_length=16, choices=PULL_POLICY, default="IfNotPresent")
    command = models.TextField(help_text="Command to run when start container", null=True, blank=True)
    args = models.TextField(
        help_text="Comma separated args to run with command\
        when instantiating container.",
        null=True,
        blank=True,
    )
    port = models.IntegerField(default=80, help_text="Port to expose.")
    volume_mounts = models.ManyToManyField("KubernetesVolumeMount", blank=True, help_text="Mounts for any number of volumes")

    def get_obj(self):
        """
        :description: Generate container spec.
        """
        return client.V1Container(
            name=self.slug,
            image=":".join([self.image_name, self.image_tag]),
            image_pull_policy=self.image_pull_policy,
            ports=[client.V1ContainerPort(container_port=self.port)],
            volume_mounts=[
                vm.get_obj() for vm in self.volume_mounts.all()
            ] if self.volume_mounts else None,
            command=[self.command] if self.command else None,
            args=self.args.split(",") if self.args else None,
        )


class KubernetesConfigMap(KubernetesMetadataObjBase):
    """
    KubernetesConfigMap
    :type: model
    :description: Holds data related to a kubernetes volume mount.
    :inherits: kubernetes_manager.models.base.KubernetesMetadataObjBase
    :fields: kind, data, binary, override_name, namespace
    """

    kind = models.CharField(max_length=16, default="ConfigMap")
    data = JSONField(default=dict, null=True, blank=True)
    binary = models.BinaryField(null=True, blank=True)
    override_name = models.CharField(max_length=32, null=True, blank=True, default="ConfigMap")
    namespace = models.ForeignKey("KubernetesNamespace", on_delete=models.CASCADE)

    def get_obj(self):
        """
        :description: Generate configmap spec.
        """
        return client.V1ConfigMap(
            metadata=client.V1ObjectMeta(name=self.slug, labels=self.labels, annotations=self.annotations),
            kind=self.kind,
            data=self.data if self.data else None,
            binary_data={str(self.override_name): self.binary} if self.binary else None,
        )

    def deploy(self):
        """
        :description: Deploy configmap obj.
        """
        api_instance = self.get_client(API=client.CoreV1Api)
        body = self.get_obj()
        api_response = api_instance.create_namespaced_config_map(body=body, namespace=self.namespace.slug)
        self.kuid = api_response.metadata.uid
        self.save()
        return str(api_response.status)

    def remove(self):
        """
        :description: Delete configmap from namespace.
        """
        api_instance = self.get_client(API=client.CoreV1Api)
        api_response = api_instance.delete_namespaced_config_map(name=self.slug, namespace=self.namespace.slug)
        self.kuid = None
        self.save()
        return str(api_response.status)


class KubernetesPodTemplate(KubernetesMetadataObjBase):
    """
    KubernetesPodTemplate
    :type: model
    :description: Holds data related to a kubernetes pod spec.
    :inherits: kubernetes_manager.models.base.KubernetesMetadataObjBase
    :fields: volume, primary_container, secondary_container, restart_policy
    """

    volumes = models.ManyToManyField("KubernetesVolume", blank=True, help_text="All volumes to be created for a pod.")
    containers = models.ManyToManyField("KubernetesContainer", help_text="All containers to be included in a pod.")
    restart_policy = models.CharField(max_length=16, choices=RESTART_POLICY, default="Never", help_text="How the pod should handle restart om case of failures")

    def get_obj(self):
        """
        :description: Generate pod spec.
        """
        if self.containers:
            return client.V1PodTemplateSpec(
                metadata=client.V1ObjectMeta(name=self.slug, labels=self.labels, annotations=self.annotations),
                spec=client.V1PodSpec(
                    volumes=[
                        v.get_obj() for v in self.volumes.all()
                     ] if self.volumes is not None else None,
                    containers=[
                        c.get_obj() for c in self.containers.all()
                    ],
                    restart_policy=self.restart_policy,
                ),
            )
        else:
            raise ValueError("Containers cannot be empty or null")


class KubernetesDeployment(KubernetesNetworkingBase, KubernetesTelemetryMixin):
    """
    KubernetesDeployment
    :type: model
    :description: Holds data related to a kubernetes deployment.
    :inherits: kubernetes_manager.models.base.KubernetesNetworkingBase,
        kubernetes_manager.models.base.KubernetesTelemetryMixin
    :fields: selector, replicas, pod_template
    """

    selector = JSONField(default=dict)
    replicas = models.IntegerField(default=1)
    pod_template = models.ForeignKey("KubernetesPodTemplate", on_delete=models.CASCADE)

    def get_obj(self):
        """
        :description: Generate Deployment spec.
        """
        return client.V1Deployment(
            api_version=self.api_version,
            kind=self.kind,
            metadata=client.V1ObjectMeta(name=self.slug, labels=self.labels, annotations=self.annotations),
            spec=client.V1DeploymentSpec(selector=client.V1LabelSelector(match_labels=self.selector), replicas=self.replicas, template=self.pod_template.get_obj()),
        )

    def deploy(self):
        """
        :description: Deploy deployment obj.
        """
        api_instance = self.get_client(API=client.AppsV1Api)
        body = self.get_obj()
        api_response = api_instance.create_namespaced_deployment(body=body, namespace=self.namespace.slug)
        ticker = 0
        while self.status().unavailable_replicas > 0:
            if ticker >= self.config.get("timeout", 60):
                raise Exception("Timeout: no replicas available after {} ticks".format(str(ticker)))
            ticker += 1
            sleep(1)
        self.kuid = api_response.metadata.uid
        self.save()
        return str(api_response.status)

    def remove(self):
        """
        :description: Remove deployment from namespace.
        """
        api_instance = self.get_client(API=client.AppsV1Api)
        api_response = api_instance.delete_namespaced_deployment(name=self.slug, namespace=self.namespace.slug)
        self.kuid = None
        self.save()
        return str(api_response.status)


class KubernetesJob(KubernetesNetworkingBase, KubernetesTelemetryMixin):
    """
    KubernetesJob
    :type: model
    :description: Holds data related to a kubernetes pod spec.
    :inherits: kubernetes_manager.models.base.KubernetesNetworkingBase,
        kubernetes_manager.models.base.KubernetesTelemetryMixin
    :fields: selector, replicas, pod_template
    """

    pod_template = models.ForeignKey("KubernetesPodTemplate", on_delete=models.CASCADE)
    backoff_limit = models.IntegerField(default=3)

    def get_obj(self):
        """
        :description: Generate job spec.
        """
        return client.V1Job(
            api_version=self.api_version,
            kind=self.kind,
            metadata=client.V1ObjectMeta(name=self.slug, labels=self.labels, annotations=self.annotations),
            spec=client.V1JobSpec(template=self.pod_template.get_obj(), backoff_limit=self.backoff_limit, ttl_seconds_after_finished=10),
        )

    def deploy(self):
        """
        :description: Deploy job to ns.
        """
        api_instance = self.get_client(API=client.BatchV1Api)
        body = self.get_obj()
        api_response = api_instance.create_namespaced_job(body=body, namespace=self.namespace.slug)
        self.kuid = api_response.metadata.uid
        self.save()
        return str(api_response.status)

    def remove(self):
        """
        :description: Remove job from ns.
        """
        api_instance = self.get_client(API=client.BatchV1Api)
        api_response = api_instance.delete_namespaced_job(name=self.slug, namespace=self.namespace.slug)
        self.kuid = None
        self.save()
        return str(api_response.status)


class KubernetesService(KubernetesNetworkingBase):
    """
    KubernetesService
    :type: model
    :description: Holds data related to a kubernetes service.
    :inherits: kubernetes_manager.models.base.KubernetesNetworkingBase
    :fields: selector, target_port
    """

    selector = JSONField(default=dict)
    target_port = models.IntegerField(default=80)

    def get_obj(self):
        """
        :description: Generate service spec.
        """
        return client.V1Service(
            api_version=self.api_version,
            kind=self.kind,
            metadata=client.V1ObjectMeta(name=self.slug, labels=self.labels, annotations=self.annotations),
            spec=client.V1ServiceSpec(
                selector=client.V1LabelSelector(match_labels=self.selector), ports=[client.V1ServicePort(port=self.port, target_port=self.target_port)]
            ),
        )

    def deploy(self):
        """
        :description: Deploy service to ns.
        """
        api_instance = self.get_client(API=client.ExtensionsV1beta1Api)
        body = self.get_obj()
        api_response = api_instance.create_namespaced_service(body=body, namespace=self.namespace.slug)
        self.kuid = api_response.metadata.uid
        self.save()
        return str(api_response.status)

    def remove(self):
        """
        :description: Remove service from ns.
        """
        api_instance = self.get_client(API=client.ExtensionsV1beta1Api)
        api_response = api_instance.delete_namespaced_service(name=self.slug, namespace=self.namespace.slug)
        self.kuid = None
        self.save()
        return str(api_response.status)


class KubernetesIngress(KubernetesNetworkingBase):
    """
    KubernetesIngress
    :type: model
    :description: Holds data related to a kubernetes ingress.
    :inherits: kubernetes_manager.models.base.KubernetesNetworkingBase
    :fields: hostname, path, target_service
    """

    hostname = models.CharField(max_length=255, default="localhost")
    path = models.CharField(max_length=255, default="/")
    target_service = models.ForeignKey("KubernetesService", on_delete=models.CASCADE)

    def get_obj(self):
        """
        :description: Generate ingress obj.
        """
        return client.NetworkingV1beta1Ingress(
            api_version=self.api_version,
            kind=self.kind,
            metadata=client.V1ObjectMeta(name=self.slug, annotations=self.annotations),
            spec=client.NetworkingV1beta1IngressSpec(
                selector=client.V1LabelSelector(match_labels=self.selector),
                rules=[
                    client.NetworkingV1beta1IngressRule(
                        host=self.hostname,
                        http=client.NetworkingV1beta1HTTPIngressRuleValue(
                            paths=[
                                client.NetworkingV1beta1HTTPIngressPath(
                                    path=self.path,
                                    backend=client.NetworkingV1beta1IngressBackend(service_port=self.target_service.port, service_name=self.target_service.slug),
                                )
                            ]
                        ),
                    )
                ],
            ),
        )

    def deploy(self):
        """
        :description: Deploy ingress to ns.
        """
        api_instance = self.get_client(API=client.ExtensionsV1beta1Api)
        body = self.get_obj()
        api_response = api_instance.create_namespaced_ingress(body=body, namespace=self.namespace.slug)
        self.kuid = api_response.metadata.uid
        self.save()
        return str(api_response.status)

    def remove(self):
        """
        :description: Remove ingress from ns.
        """
        api_instance = self.get_client(API=client.ExtensionsV1beta1Api)
        api_response = api_instance.delete_namespaced_ingress(name=self.slug, namespace=self.namespace.slug)
        self.kuid = None
        self.save()
        return str(api_response.status)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import inspect
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import django
from django.utils.encoding import force_text
from django.utils.html import strip_tags

sys.path.insert(0, os.path.abspath("../.."))
os.environ["DJANGO_SETTINGS_MODULE"] = "_django.settings"
django.setup()


# -- Project information -----------------------------------------------------

project = "Django Kubernetes Manager"
copyright = "2020, Bradley Reimers"
author = "Bradley Reimers"
master_doc = "index"

# The full version, including alpha/beta/rc tags
release = "0.1.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.doctest", "sphinx.ext.coverage", "sphinxcontrib.httpdomain", "sphinxcontrib.openapi"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

autodoc_mock_imports = ["django", "kubernetes", "rest_framework", "django_extensions"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

def process_docstring(app, what, name, obj, options, lines):
    # This causes import errors if left outside the function
    from django.db import models

    # Only look at objects that inherit from Django's base model class
    if inspect.isclass(obj) and issubclass(obj, models.Model):
        # Grab the field list from the meta class
        fields = obj._meta.get_fields()

        for field in fields:
            # Skip ManyToOneRel and ManyToManyRel fields which have no 'verbose_name' or 'help_text'
            if not hasattr(field, 'verbose_name'):
                continue

            # Decode and strip any html out of the field's help text
            help_text = strip_tags(force_text(field.help_text))

            # Decode and capitalize the verbose name, for use if there isn't
            # any help text
            verbose_name = force_text(field.verbose_name).capitalize()

            if help_text:
                # Add the model field to the end of the docstring as a param
                # using the help text as the description
                lines.append(u':param %s: %s' % (field.attname, help_text))
            else:
                # Add the model field to the end of the docstring as a param
                # using the verbose name as the description
                lines.append(u':param %s: %s' % (field.attname, verbose_name))

            lines.append(u':type %s: %s' % (field.attname, type(field).__name__))

    # Return the extended docstring
    return lines

def setup(app):
    # Register the docstring processor with sphinx
    app.connect('autodoc-process-docstring', process_docstring)



from .base import (
    KubernetesConfigMapSerializer,
    KubernetesContainerSerializer,
    KubernetesDeploymentSerializer,
    KubernetesIngressSerializer,
    KubernetesJobSerializer,
    KubernetesNamespaceSerializer,
    KubernetesPodTemplateSerializer,
    KubernetesServiceSerializer,
    KubernetesVolumeMountSerializer,
    KubernetesVolumeSerializer,
    TargetClusterSerializer,
)



from kubernetes_manager.models import (
    KubernetesBase,
    KubernetesConfigMap,
    KubernetesContainer,
    KubernetesDeployment,
    KubernetesIngress,
    KubernetesJob,
    KubernetesMetadataObjBase,
    KubernetesNamespace,
    KubernetesNetworkingBase,
    KubernetesPodTemplate,
    KubernetesService,
    KubernetesVolume,
    KubernetesVolumeMount,
    TargetCluster,
)
from rest_framework import serializers


class TargetClusterSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = TargetCluster
        fields = ["title", "api_endpoint", "telemetry_endpoint", "config"]


class KubernetesBaseSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = KubernetesBase
        fields = ["title", "description", "cluster", "config"]
        abstract = True


class KubernetesMetadataObjBaseSerializer(KubernetesBaseSerializer):
    class Meta:
        model = KubernetesMetadataObjBase
        fields = KubernetesBaseSerializer.Meta.fields + ["labels", "annotations"]
        abstract = True


class KubernetesNetworkingBaseSerializer(KubernetesMetadataObjBaseSerializer):
    class Meta:
        model = KubernetesNetworkingBase
        fields = KubernetesMetadataObjBaseSerializer.Meta.fields + ["api_version", "kind", "port", "namespace", "kuid"]
        abstract = True


class KubernetesVolumeSerializer(KubernetesBaseSerializer):
    class Meta:
        model = KubernetesVolume
        fields = KubernetesBaseSerializer.Meta.fields


class KubernetesVolumeMountSerializer(KubernetesBaseSerializer):
    class Meta:
        model = KubernetesVolumeMount
        fields = KubernetesBaseSerializer.Meta.fields + ["mount_path", "sub_path"]


class KubernetesNamespaceSerializer(KubernetesMetadataObjBaseSerializer):
    class Meta:
        model = KubernetesNamespace
        fields = KubernetesMetadataObjBaseSerializer.Meta.fields + ["api_version", "kind", "exists"]


class KubernetesConfigMapSerializer(KubernetesMetadataObjBaseSerializer):
    class Meta:
        model = KubernetesConfigMap
        fields = KubernetesMetadataObjBaseSerializer.Meta.fields + [
            "data",
            "kind",
        ]


class KubernetesContainerSerializer(KubernetesBaseSerializer):
    class Meta:
        model = KubernetesContainer
        fields = KubernetesBaseSerializer.Meta.fields + ["image_name", "image_tag", "image_pull_policy", "command", "args", "port", "volume_mounts"]


class KubernetesPodTemplateSerializer(KubernetesMetadataObjBaseSerializer):
    class Meta:
        model = KubernetesPodTemplate
        fields = KubernetesMetadataObjBaseSerializer.Meta.fields + ["volumes", "containers", "restart_policy"]


class KubernetesDeploymentSerializer(KubernetesNetworkingBaseSerializer):
    class Meta:
        model = KubernetesDeployment
        fields = KubernetesNetworkingBaseSerializer.Meta.fields + ["selector", "replicas", "pod_template"]


class KubernetesJobSerializer(KubernetesNetworkingBaseSerializer):
    class Meta:
        model = KubernetesJob
        fields = KubernetesNetworkingBaseSerializer.Meta.fields + ["backoff_limit", "pod_template"]


class KubernetesServiceSerializer(KubernetesNetworkingBaseSerializer):
    class Meta:
        model = KubernetesService
        fields = KubernetesNetworkingBaseSerializer.Meta.fields + ["selector", "target_port"]


class KubernetesIngressSerializer(KubernetesNetworkingBaseSerializer):
    class Meta:
        model = KubernetesIngress
        fields = KubernetesNetworkingBaseSerializer.Meta.fields + ["hostname", "path", "target_service"]



import json
import re

from django.db import models
from kubernetes_manager.consts import byte_units
from kubernetes import client, config


class KubernetesTelemetryMixin(models.Model):
    """
    KubernetesTelemetryMixin
    :type: mixin
    :description: Extends child model to include telemetry features.
    :inherits: django.db.models.Model
    :fields: object_status, average_cpu_usage,
        average_mem_usage, cpu_usage_seconds, mem_usage_seconds
    """

    object_status = models.CharField(max_length=128, null=True, blank=True, help_text="status of the object in Kubernetes")
    average_cpu_usage = models.DecimalField(null=True, blank=True, max_digits=8, decimal_places=4, help_text="Average PIT CPU units consumed")
    average_mem_usage = models.IntegerField(null=True, blank=True, help_text="Average PIT bytes consumed")
    cpu_usage_seconds = models.DecimalField(null=True, blank=True, max_digits=8, decimal_places=4, help_text="Average cpu usage * seconds live")
    mem_usage_seconds = models.IntegerField(null=True, blank=True, help_text="Average mem usage * seconds live")

    class Meta:
        abstract = True

    def splitNumeric(self, size):
        """
        :description: Parses string into numeric component.
        """
        return filter(None, re.split(r"(\d+)", size))

    def parseSize(self, size):
        """
        :description: Parses string as numeric, suffix and converts to bytes.
        """
        number, unit = [string for string in self.splitNumeric(size)]
        return int(float(number) * byte_units[unit])

    def read_pod_metrics(self):
        """
        :description: Uses metrics_server to get  cadvisor data.
        """
        api_instance = self.get_client(API=client.CustomObjectsApi)
        pod_name = self.slug
        pod_namespace = self.namespace.slug
        items = api_instance.list_cluster_custom_object("metrics.k8s.io", "v1beta1", "pods").get("items", [])
        return [pod for pod in items if pod_name in pod.get("metadata", {}).get("name") and pod_namespace in pod.get("metadata", {}).get("namespace")]

    def read_pod_usage(self):
        """
        :description: Converts metrics into dictionary for api usage.
        """
        pod_name = self.pod_template.slug
        pod_namespace = self.namespace.slug
        pod_metrics = self.read_pod_metrics()
        cpu = 0.000
        memory = 0
        for metric in pod_metrics:
            for container in metric.get("containers", []):
                ccpu = container.get("usage", {}).get("cpu", None)
                cmem = container.get("usage", {}).get("memory", None)
                if "m" in ccpu:
                    ccpu = int(ccpu.split("m")[0]) / 1000.000
                else:
                    ccpu = int(ccpu)
                cpu += ccpu
                memory += self.parseSize(cmem)
        return {"cpu": cpu, "memory": memory}

    def status(self):
        """
        :description: Returns status data of object.
        """
        type = self._meta.model_name
        name = self.slug
        namespace = self.namespace.slug
        api_instance = self.get_client(API=client.ExtensionsV1beta1Api)
        if type == "kubernetesjob":
            api_response = api_instance.read_namespaced_job_status(name, namespace)
        if type == "kubernetesdeployment":
            api_response = api_instance.read_namespaced_deployment_status(name, namespace)
        return api_response.status



from .base import KubernetesBase, KubernetesMetadataObjBase, KubernetesNetworkingBase
from .kube import (
    KubernetesConfigMap,
    KubernetesContainer,
    KubernetesDeployment,
    KubernetesIngress,
    KubernetesJob,
    KubernetesNamespace,
    KubernetesPodTemplate,
    KubernetesService,
    KubernetesVolume,
    KubernetesVolumeMount,
)
from .target_cluster import TargetCluster



import json
from tempfile import NamedTemporaryFile
from uuid import uuid4

from django.contrib.postgres.fields import JSONField
from django.db import models
from django_extensions.db.models import TitleSlugDescriptionModel
from kubernetes import client, config


class KubernetesBase(TitleSlugDescriptionModel):
    """
    KubernetesBase
    :type: model (abstract)
    :description: Base parent model that all subsequent models inherit from.
    :inherits: django_extensions.db.models.TitleSlugDescriptionModel
    :fields: id, cluster, config, deployed, deleted
    """

    id = models.UUIDField(default=uuid4, editable=False, primary_key=True, help_text="UUID Auto field.")
    cluster = models.ForeignKey("TargetCluster", on_delete=models.SET_NULL, null=True, help_text="ForeignKey to TargetCluster object.")
    config = JSONField(default=dict, null=True, blank=True, help_text="Pass in extra parameters here.")
    deployed = models.DateTimeField(null=True, blank=True, help_text="Time when object is applied to cluster.")
    removed = models.DateTimeField(null=True, blank=True, help_text="Time when object is removed from cluster.")

    class Meta:
        abstract = True

    def slugify_function(self, content):
        """
        :description: Overrides default slugify with custom logic.
        """
        return self.title.replace("_", "-").replace(" ", "-").lower()

    def get_client(self, API=client.CoreV1Api, **kwargs):
        """Gets a k8s api client

            Args:
                API (client.<type>) - Kubernetes Client Type
            Returns:
                object of type <API>
        """

        if "persist_config" not in kwargs:
            kwargs["persist_config"] = False
        with NamedTemporaryFile() as ntf:
            kwargs["config_file"] = ntf.name
            cc = json.dumps(self.cluster.config) if isinstance(self.cluster.config, dict) else self.cluster.config
            with open(ntf.name, "w") as f:
                f.write(cc)
            return API(api_client=config.new_client_from_config(config_file=ntf.name))


class KubernetesMetadataObjBase(KubernetesBase):
    """
    KubernetesMetadataObjBase
    :type: model (abstract)
    :description: Extends KubernetesBase to include metadata fields.
    :inherits: kubernetes_manager.models.base.KubernetesBase
    :fields: labels, annotations
    """

    labels = JSONField(default=dict, help_text="Dictionary store equivalent to Labels in Kubernetes API")
    annotations = JSONField(default=dict, null=True, blank=True, help_text="Dictionary store equivalent to Annotations in Kubernetes API")

    class Meta:
        abstract = True


class KubernetesNetworkingBase(KubernetesMetadataObjBase):
    """
    KubernetesNetworkingBase
    :type: model (abstract)
    :description: Extends KubernetesMetadataObjBase to include network fields.
    :inherits: kubernetes_manager.models.base.KubernetesMetadataObjBase
    :fields: labels, annotations
    """

    api_version = models.CharField(max_length=16, default="v1", help_text="API version used to deploy child object.")
    kind = models.CharField(max_length=16, help_text="String representation of Kubernetes object kind")
    port = models.IntegerField(default=80, help_text="Port object will expose")
    namespace = models.ForeignKey("KubernetesNamespace", on_delete=models.CASCADE, help_text="Live namespace the object is associated with.")
    kuid = models.CharField(max_length=48, null=True, blank=True, help_text="Object's UID in the cluster")

    class Meta:
        abstract = True



from rest_framework import permissions, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from ..models import (
    KubernetesConfigMap,
    KubernetesContainer,
    KubernetesDeployment,
    KubernetesIngress,
    KubernetesJob,
    KubernetesNamespace,
    KubernetesPodTemplate,
    KubernetesService,
    KubernetesVolume,
    KubernetesVolumeMount,
    TargetCluster,
)
from ..serializers import (
    KubernetesConfigMapSerializer,
    KubernetesContainerSerializer,
    KubernetesDeploymentSerializer,
    KubernetesIngressSerializer,
    KubernetesJobSerializer,
    KubernetesNamespaceSerializer,
    KubernetesPodTemplateSerializer,
    KubernetesServiceSerializer,
    KubernetesVolumeMountSerializer,
    KubernetesVolumeSerializer,
    TargetClusterSerializer,
)


class TargetClusterViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows cluster configs to be edited or deleted
    """

    queryset = TargetCluster.objects.all()
    serializer_class = TargetClusterSerializer
    permission_classes = [permissions.IsAuthenticated]


class KubernetesNamespaceViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows namespaces to be created or deleted
    """

    queryset = KubernetesNamespace.objects.all()
    serializer_class = KubernetesNamespaceSerializer
    permission_classes = [permissions.IsAuthenticated]

    @action(methods=["get"], detail=True)
    def deploy(self, request, *args, **kwargs):
        """
        Action to deploy the namespace resource to target cluster.
        """
        return Response(KubernetesNamespace.objects.get(pk=kwargs["pk"]).deploy())

    @action(methods=["get"], detail=True)
    def remove(self, request, *args, **kwargs):
        """
        Action to delete the kubernetes namespace from the cluster.
        """
        return Response(KubernetesNamespace.objects.get(pk=kwargs["pk"]).remove())


class KubernetesConfigMapViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows configmaps to be edited or deleted
    """

    queryset = KubernetesConfigMap.objects.all()
    serializer_class = KubernetesConfigMapSerializer
    permission_classes = [permissions.IsAuthenticated]

    @action(methods=["get"], detail=True)
    def deploy(self, request, *args, **kwargs):
        """
        Action to deploy the ConfigMap resource to target cluster.
        """
        return Response(KubernetesConfigMap.objects.get(pk=kwargs["pk"]).deploy())

    @action(methods=["get"], detail=True)
    def remove(self, request, *args, **kwargs):
        """
        Action to delete the kubernetes ConfigMap from the cluster.
        """
        return Response(KubernetesConfigMap.objects.get(pk=kwargs["pk"]).remove())


class KubernetesVolumeViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows volumes to be edited or deleted.
    """

    queryset = KubernetesVolume.objects.all()
    serializer_class = KubernetesVolumeSerializer
    permission_classes = [permissions.IsAuthenticated]


class KubernetesVolumeMountViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows volumes to be edited or deleted.
    """

    queryset = KubernetesVolumeMount.objects.all()
    serializer_class = KubernetesVolumeMountSerializer
    permission_classes = [permissions.IsAuthenticated]


class KubernetesContainerViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows containers to be edited or deleted.
    """

    queryset = KubernetesContainer.objects.all()
    serializer_class = KubernetesContainerSerializer
    permission_classes = [permissions.IsAuthenticated]


class KubernetesPodTemplateViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows pod templates to be edited or deleted.
    """

    queryset = KubernetesPodTemplate.objects.all()
    serializer_class = KubernetesPodTemplateSerializer
    permission_classes = [permissions.IsAuthenticated]


class KubernetesDeploymentViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows deployments to be edited or deleted.
    """

    queryset = KubernetesDeployment.objects.all()
    serializer_class = KubernetesDeploymentSerializer
    permission_classes = [permissions.IsAuthenticated]

    @action(methods=["get"], detail=True)
    def deploy(self, request, *args, **kwargs):
        """
        Action to deploy the kubernetes resource to target cluster.
        """
        return Response(KubernetesDeployment.objects.get(pk=kwargs["pk"]).deploy())

    @action(methods=["get"], detail=True)
    def remove(self, request, *args, **kwargs):
        """
        Action to delete the kubernetes resource from the cluster/namespaces
        """
        return Response(KubernetesDeployment.objects.get(pk=kwargs["pk"]).remove())

    @action(methods=["get"], detail=True)
    def pod_usage(self, request, *args, **kwargs):
        """
        Action to fetch point-in-time cpu and memory usage of pod.
        """
        return Response(KubernetesDeployment.objects.get(pk=kwargs["pk"]).read_pod_usage())


class KubernetesServiceViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows services to be edited or deleted.
    """

    queryset = KubernetesService.objects.all()
    serializer_class = KubernetesServiceSerializer
    permission_classes = [permissions.IsAuthenticated]

    @action(methods=["get"], detail=True)
    def deploy(self, request, *args, **kwargs):
        """
        Action to deploy the kubernetes resource to target cluster.
        """
        return Response(KubernetesService.objects.get(pk=kwargs["pk"]).deploy())

    @action(methods=["get"], detail=True)
    def remove(self, request, *args, **kwargs):
        """
        Action to delete the kubernetes resource from the cluster/namespace.
        """
        return Response(KubernetesService.objects.get(pk=kwargs["pk"]).remove())


class KubernetesIngressViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows ingress to be edited or deleted.
    """

    queryset = KubernetesIngress.objects.all()
    serializer_class = KubernetesIngressSerializer
    permission_classes = [permissions.IsAuthenticated]

    @action(methods=["get"], detail=True)
    def deploy(self, request, *args, **kwargs):
        """
        Action to deploy the kubernetes resource to target cluster.
        """
        return Response(KubernetesIngress.objects.get(pk=kwargs["pk"]).deploy())

    @action(methods=["get"], detail=True)
    def remove(self, request, *args, **kwargs):
        """
        Action to delete the kubernetes resource from the cluster/namespace.
        """
        return Response(KubernetesIngress.objects.get(pk=kwargs["pk"]).remove())


class KubernetesJobViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows jobs to be edited or deleted.
    """

    queryset = KubernetesJob.objects.all()
    serializer_class = KubernetesJobSerializer
    permission_classes = [permissions.IsAuthenticated]

    @action(methods=["get"], detail=True)
    def deploy(self, request, *args, **kwargs):
        """
        Action to deploy the kubernetes resource to target cluster.
        """
        return Response(KubernetesJob.objects.get(pk=kwargs["pk"]).deploy())

    @action(methods=["get"], detail=True)
    def remove(self, request, *args, **kwargs):
        """
        Action to delete the kubernetes resource from the target cluster/ns.
        """
        return Response(KubernetesJob.objects.get(pk=kwargs["pk"]).remove())

    @action(methods=["get"], detail=True)
    def pod_usage(self, request, *args, **kwargs):
        """
        Action to fetch point-in-time cpu and memory usage of pod.
        """
        return Response(KubernetesJob.objects.get(pk=kwargs["pk"]).read_pod_usage())



import random
import string

from django.test import TestCase

from .factories import (
    KubernetesConfigMapFactory,
    KubernetesContainerFactory,
    KubernetesDeploymentFactory,
    KubernetesJobFactory,
    KubernetesNamespaceFactory,
    KubernetesPodTemplateFactory,
    KubernetesVolumeFactory,
    KubernetesVolumeMountFactory,
    TargetClusterFactory,
)


class TargetClusterTestCase(TestCase):
    def test_model(self):
        pass

    def test_api(self):
        pass


class KubernetesNamespaceTestCase(TestCase):
    def test_model(self):
        obj = KubernetesNamespaceFactory()
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesNamespaceFactory()
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def namespace_create_api(self):
        arf = APIRequestFactory()
        req = arf.post(
            "/namespaces/",
            {
                "title": str(factory.fuzzy.FuzzyText()),
                "description": str(factory.fuzzy.FuzzyText()),
                "cluster": "http://127.0.0.1:8000/dkm/api/clusters/1/",
                "labels": {"app": "test"},
                "annotations": {"type": "project"},
                "api_version": "v1",
                "kind": "Namespace",
                "exists": False,
            },
        )

    def namespace_deploy_api(self):
        arf = APIRequestFactory()
        req = arf.get("/namespaces/{}/deploy".format(id))

    def namespace_remove_api(self):
        arf = APIRequestFactory()
        req = arf.get("/namespaces/{}/remove".format(id))

    def namespace_delete_api(self):
        arf = APIRequestFactory()
        req = arf.delete("/namespaces/{}/".format(id))


class KubernetesVolumeTestCase(TestCase):
    def test_model(self):
        obj = KubernetesVolumeFactory()
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesVolumeFactory()
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def test_api(self):
        pass


class KubernetesVolumeMountTestCase(TestCase):
    def test_model(self):
        obj = KubernetesVolumeMountFactory()
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesVolumeMountFactory()
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def test_api(self):
        pass


class KubernetesConfigMapTestCase(TestCase):
    def test_model(self):
        obj = KubernetesConfigMapFactory()
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesConfigMapFactory()
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def test_api(self):
        pass


class KubernetesContainerTestCase(TestCase):
    def test_model(self):
        obj = KubernetesContainerFactory()
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesContainerFactory()
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def test_api(self):
        pass


class KubernetesPodTemplateTestCase(TestCase):
    def test_model(self):
        rstr = ''.join(random.choices(string.ascii_lowercase + string.digits, k = 7))
        obj = KubernetesPodTemplateFactory(
            containers=(
                KubernetesContainerFactory(title=rstr), KubernetesContainerFactory(title=rstr+"-2")
            ),
            volumes=(
                KubernetesVolumeFactory(title=rstr), KubernetesVolumeFactory(title=rstr+"-2")
            )
        )
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesPodTemplateFactory(
            containers=(
                KubernetesContainerFactory(), KubernetesContainerFactory()
            )
        )
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def test_api(self):
        pass


class KubernetesDeploymentTestCase(TestCase):
    def test_model(self):
        obj = KubernetesDeploymentFactory()
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesDeploymentFactory()
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def test_api(self):
        pass


class KubernetesJobTestCase(TestCase):
    def test_model(self):
        obj = KubernetesJobFactory()
        self.assertIsNotNone(obj)
        self.assertIsNotNone(obj.pk)

    def test_get_obj(self):
        obj = KubernetesJobFactory()
        obj = obj.get_obj()
        self.assertIsNotNone(obj)

    def test_api(self):
        pass
